//
// Created by philipp on 29.12.17.
//

#ifndef NETWORKINGLIB_ASYNCDATAGRAMRECEIVER_H
#define NETWORKINGLIB_ASYNCDATAGRAMRECEIVER_H

#include "Networking.h"
#include "Utils.h"

namespace networking
{
namespace message
{

template<typename Message>
class DatagramReceiver : public std::enable_shared_from_this<DatagramReceiver<Message>>
{
private:
    struct PrivateTag
    {
    };

public:
    using Ptr = std::shared_ptr<DatagramReceiver<Message>>;

    using Buffer = typename Message::Buffer;

    using Endpoint = boost::asio::ip::udp::endpoint;
    using Socket = boost::asio::ip::udp::socket;

    using ReceiveHandler = std::function<
        void(const Endpoint & senderEndpoint, const Message & message, const boost::system::error_code & error)>;

    DatagramReceiver(PrivateTag, Networking & net, const std::string & host, std::uint16_t port)
        : net(net)
          , bindingEndpoint(boost::asio::ip::address::from_string(host), port)
          , socket(net.getIoService(), bindingEndpoint)
          , buffer(Message::emptyBuffer())
    {}

    static Ptr create(Networking & net, const std::string & host, std::uint16_t port)
    {
        return std::make_shared<DatagramReceiver<Message>>(PrivateTag{}, net, host, port);
    }

    DatagramReceiver(const DatagramReceiver &) = delete;

    DatagramReceiver & operator=(const DatagramReceiver &) = delete;

    DatagramReceiver(DatagramReceiver &&) = delete;

    DatagramReceiver & operator=(DatagramReceiver &&) = delete;

    bool isReceiving() const
    { return receiving; };

    bool asyncReceive(ReceiveHandler receiveHandler, const time::Duration & timeout)
    {
        if (receiving)
            return false;

        receiving = true;
        if (!socket.is_open())
            socket = Socket(net.getIoService(), bindingEndpoint);

        auto self = this->shared_from_this();
        networking::utils::asyncReceiveFrom(
            net, socket, boost::asio::buffer(buffer), senderEndpoint,
            [self, receiveHandler](const boost::system::error_code & receiveError,
                                   std::size_t bytesTransferred)
            {
                Message msg{self->buffer};

                boost::system::error_code error = receiveError;
                if (!receiveError && bytesTransferred < self->buffer.size())
                    error = boost::asio::error::message_size;

                auto senderEndpoint = self->senderEndpoint;
                self->receiving = false;
                receiveHandler(senderEndpoint, msg, error);
            }, timeout);

        return true;
    }

    void stopReceiving()
    {
        if (!receiving)
            return;

        socket.close();
    }

private:
    Networking & net;
    Endpoint bindingEndpoint;
    Socket socket;
    Buffer buffer;
    Endpoint senderEndpoint;
    std::atomic<bool> receiving{false};
};

}
}

#endif //NETWORKINGLIB_ASYNCDATAGRAMRECEIVER_H
