//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_ProtocolNETWORKSERVICESERVER_H
#define PROTOCOL_ProtocolNETWORKSERVICESERVER_H

#include "Message.h"
#include "Networking.h"

namespace networking
{
namespace service
{

template<typename Service>
class Server : public std::enable_shared_from_this<Server<Service>>
{
private:
    struct PrivateTag
    {
    };

public:
    using RequestMessage = typename Service::RequestMessage;
    using ResponseMessage = typename Service::ResponseMessage;

    using Ptr = std::shared_ptr<Server<Service>>;

    using Endpoint = typename boost::asio::ip::tcp::endpoint;
    using Socket = typename boost::asio::ip::tcp::socket;
    using Acceptor = typename boost::asio::ip::tcp::acceptor;

    using RequestReceivedHandler =
    std::function<ResponseMessage(const Endpoint & clientEndpoint,
                                  const RequestMessage & requestMessage)>;

    static Ptr create(Networking & net,
                      const std::string & host,
                      uint16_t port,
                      RequestReceivedHandler requestReceivedHandler)
    {
        return std::make_shared<Server<Service>>(PrivateTag{}, net, host, port, requestReceivedHandler);
    }

    // Should not be used outside.
    Server(PrivateTag,
           Networking & net,
           const std::string & host,
           uint16_t port,
           RequestReceivedHandler requestReceivedHandler)
        : net(net)
          , requestReceivedHandler(requestReceivedHandler)
          , bindingEndpoint(boost::asio::ip::address::from_string(host), port)
          , acceptor(net.getIoService(), bindingEndpoint)
          , socket(net.getIoService())
    {}

    Server(const Server &) = delete;

    Server & operator=(const Server &) = delete;

    Server(Server &&) = delete;

    Server & operator=(Server &&) = delete;

    void advertiseService()
    {
        if (running)
            return;

        running = true;
        accept();
    }

    void stop()
    {
        running = false;
        socket.close();
        acceptor.close();
    }

private:
    Networking & net;
    RequestReceivedHandler requestReceivedHandler;
    Endpoint bindingEndpoint;
    Acceptor acceptor;
    Socket socket;
    std::atomic<bool> running{false};

    void accept()
    {
        if (!socket.is_open())
            socket = Socket(net.getIoService());

        if (!acceptor.is_open())
            acceptor = Acceptor(net.getIoService(), bindingEndpoint);

        auto self = this->shared_from_this();
        acceptor.async_accept(socket, [self](boost::system::error_code acceptError)
        {
            utils::SocketCloser<Socket> socketCloser(self->socket);

            if (!self->running)
                return;

            if (!acceptError)
            {
                try
                {
                    using namespace std::chrono_literals;
                    auto request = message::receive<RequestMessage>(self->net, self->socket, 3s);
                    auto response = self->requestReceivedHandler(self->socket.remote_endpoint(), request);
                    message::send(self->net, self->socket, response, 3s);
                }
                catch (const boost::system::system_error & error)
                {
                    // If a receive has timed out we treat it like we've never received any message (and therefor we
                    // do not call the handler). We cannot be sure that the message is going to be received at the other side anyway,
                    // so we don't handle anything sending-wise.
                }
            }

            // The next accept event will be put on the event queue.
            self->accept();
        });
    }
};

}
}

#endif //PROTOCOL_ProtocolNETWORKSERVICESERVER_H
