//
// Created by philipp on 02.12.17.
//

#ifndef PROTOCOL_NETWORKUTILS_H
#define PROTOCOL_NETWORKUTILS_H

#include <boost/asio/ip/tcp.hpp>
#include <boost/asio.hpp>
#include "Networking.h"
#include "Timer.h"

namespace networking
{
namespace utils
{

template<typename AsyncOperation, typename Socket, typename... Args>
void timedAsyncSocketOperation(Networking & net,
                               AsyncOperation asyncOperation,
                               Socket & socket,
                               const time::Duration & timeout,
                               Args && ... args)
{
    auto & ioService = net.getIoService();

    auto timer = time::Timer::create(net);
    timer->startTimeout(
        timeout,
        [&socket]()
        {
            // Timeout expired: close the socket.
            boost::system::error_code ignoredError;
            socket.close(ignoredError);
        });

    // A 'would_block' socketError is guaranteed to never occur on an asynchronous operation.
    boost::system::error_code socketError = boost::asio::error::would_block;

    // Run asynchronous connect.
    asyncOperation(
        socket, std::forward<Args>(args)...,
        [&socketError, timer](const boost::system::error_code & error, auto ...)
        {
            timer->disable();
            // update socketError variable
            socketError = error;
        });

    // Wait until "something happens" with the socket.
    net.waitUntil([&socketError]()
                  { return socketError != boost::asio::error::would_block; });

    // Determine whether a connection was successfully established.
    // Even though our timer handler might have run to close the socket, the connect operation
    // might have notionally succeeded!
    if (socketError || !socket.is_open())
    {
        throw boost::system::system_error(
            socketError ? socketError : boost::asio::error::operation_aborted);
    }
}

template<typename Socket, typename EndpointIterator>
void connect(Networking & net,
             Socket & socket,
             const EndpointIterator & endpointIterator,
             const time::Duration & timeout)
{
    // This lambda is used because we cannot pass boost::asio::async_connect directly as a function
    // since it is a template.
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_connect(std::forward<decltype(args)>(args)...); };
    timedAsyncSocketOperation(net, asyncOperation, socket, timeout, endpointIterator);
}

template<typename SyncWriteStream, typename ConstBufferSequence>
void write(Networking & net,
           SyncWriteStream & stream,
           const ConstBufferSequence & buffer,
           const time::Duration & timeout)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_write(std::forward<decltype(args)>(args)...); };
    timedAsyncSocketOperation(net, asyncOperation, stream, timeout, buffer);
};

template<typename SyncReadStream, typename MutableBufferSequence>
void read(Networking & net,
          SyncReadStream & stream,
          const MutableBufferSequence & buffer,
          const time::Duration & timeout)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_read(std::forward<decltype(args)>(args)...); };
    timedAsyncSocketOperation(net, asyncOperation, stream, timeout, buffer);
};

template<
    typename DatagramSocket,
    typename MutableBuffers,
    typename Endpoint,
    typename ReceiveHandler>
void asyncReceiveFrom(Networking & net,
                      DatagramSocket & socket,
                      const MutableBuffers & buffer,
                      Endpoint & endpoint,
                      ReceiveHandler receiveHandler,
                      const time::Duration & timeout)
{
    auto timer = time::Timer::create(net);
    timer->startTimeout(
        timeout,
        [&socket]()
        {
            boost::system::error_code ignoredError;
            socket.close(ignoredError);
        });

    socket.async_receive_from(
        buffer, endpoint,
        [&socket, timer, receiveHandler](const boost::system::error_code & receiveError,
                                         std::size_t bytesTransferred)
        {
            timer->disable();
            auto error = receiveError;
            if (!socket.is_open())
                error = boost::asio::error::operation_aborted;
            receiveHandler(error, bytesTransferred);
        });
}

template<typename Socket>
class SocketCloser
{
public:
    SocketCloser(Socket & socket) : socket(socket)
    {}

    ~SocketCloser()
    { socket.close(); }

private:
    Socket & socket;
};

}
}

#endif //PROTOCOL_NETWORKUTILS_H
