//
// Created by philipp on 02.12.17.
//

#include "Test.h"
#include "TestUtils.h"
#include <boost/asio/ip/tcp.hpp>
#include <iostream>
#include "../include/ServiceServer.h"
#include "PlatoonService.h"
#include "../include/ServiceClient.h"
#include "../include/DatagramReceiver.h"
#include "../include/DatagramSender.h"

using boost::asio::ip::tcp;

using namespace networking;
using namespace std::chrono_literals;

void test::testServices()
{
    using namespace protocol;
    Networking net1;
    Networking net2;

    auto server = service::Server<PlatoonService>::create(
        net1, "127.0.0.1", 10001,
        [](const tcp::endpoint & clientEndpoint,
           const PlatoonMessage & requestMessage) -> PlatoonMessage
        {
            std::cout << "Request from "
                      << (int) requestMessage.getVehicleId()
                      << " with type: "
                      << (int) requestMessage.getMessageType()
                      << "\n";
            return PlatoonMessage::acceptResponse(1, 42);
        });

    server->advertiseService();

    sleep(1);

    service::Client<PlatoonService> client{net2};
    for (int i = 0; i < 5; i++)
    {
        auto response = client.call("127.0.0.1", 10001, PlatoonMessage::followerRequest(2), 5s);
        std::cout << "Response from " << (int) response.getVehicleId() << " with type: "
                  << (int) response.getMessageType()
                  << " and platoonId: " << (int) response.getPlatoonId() << "\n";
    }
}

void test::testTcpClientTimeout()
{
    using namespace protocol;
    Networking net1;
    Networking net2;

    const auto timeout = 3s;

    auto server = service::Server<PlatoonService>::create(
        net1,
        "127.0.0.1", 10001,
        [](const tcp::endpoint & clientEndpoint,
           const PlatoonMessage & requestMessage) -> PlatoonMessage
        {
            // Just sleep for 5 seconds.
            sleep(5);
            return PlatoonMessage::acceptResponse(1, 42);
        });

    server->advertiseService();
    sleep(1);

    service::Client<PlatoonService> client{net2};
    auto startTime = boost::posix_time::microsec_clock::local_time();
    try
    {
        auto response = client.call("127.0.0.1", 10001, PlatoonMessage::followerRequest(2), timeout);
        std::cout << "Response: " << response.getPlatoonId() << "\n";
        std::cout << "FAILED!";
    }
    catch (const std::exception & e)
    {
        auto nowTime = boost::posix_time::microsec_clock::local_time();
        auto timeSpend = nowTime - startTime;
        if (timeSpend.seconds() >= 2)
            std::cout << "SUCCESS!\n";
    }
}

void test::testMultipleConnections()
{
    using namespace protocol;
    Networking net1;
    Networking net2;
    Networking net3;

    auto server1 = service::Server<PlatoonService>::create(
        net1,
        "127.0.0.1", 10001,
        [](const tcp::endpoint & clientEndpoint,
           const PlatoonMessage & requestMessage) -> PlatoonMessage
        {
            return PlatoonMessage::acceptResponse(1, 42);
        });

    auto server2 = service::Server<PlatoonService>::create(
        net2,
        "127.0.0.1", 10002,
        [](const tcp::endpoint & clientEndpoint,
           const PlatoonMessage & requestMessage) -> PlatoonMessage
        {
            return PlatoonMessage::acceptResponse(2, 43);
        });

    server1->advertiseService();
    server2->advertiseService();

    sleep(1);

    service::Client<PlatoonService> client{net3};

    auto response = client.call("127.0.0.1", 10001, PlatoonMessage::followerRequest(1), 5s);
    std::cout << "Response from " << response.getVehicleId() << std::endl;

    response = client.call("127.0.0.1", 10002, PlatoonMessage::followerRequest(2), 5s);
    std::cout << "Response from " << response.getVehicleId() << std::endl;
}

void test::testStoppingServiceServer()
{
    using namespace protocol;
    Networking net1;
    Networking net2;

    auto server = service::Server<PlatoonService>::create(
        net1,
        "127.0.0.1", 10001,
        [](const tcp::endpoint & clientEndpoint,
           const PlatoonMessage & requestMessage) -> PlatoonMessage
        {
            // Just sleep for 3 seconds.
            sleep(3);
            return PlatoonMessage::acceptResponse(1, 42);
        });

    sleep(1);

    service::Client<PlatoonService> client{net2};
    try
    {
        client.call("127.0.0.1", 10001, PlatoonMessage::followerRequest(42), 1s);
    }
    catch (const boost::system::system_error & error)
    {
        assert(error.code() == boost::asio::error::operation_aborted);
        server->stop();
        server->advertiseService();
        auto response = client.call("127.0.0.1", 10001, PlatoonMessage::followerRequest(42), 5s);
        if (response.getMessageType() == messageTypes::ACCEPT_RESPONSE)
            std::cout << "SUCCESS!\n";
    }
}

void test::testAsyncDatagramReceiver()
{
    using namespace protocol;
    Networking net1;
    Networking net2;

    auto receiver = message::DatagramReceiver<PlatoonMessage>::create(net1, "127.0.0.1", 10000);

    message::DatagramSender<PlatoonMessage> sender{net2};

    receiver->asyncReceive(
        [](const boost::asio::ip::udp::endpoint & senderEndpoint,
           const PlatoonMessage & message,
           boost::system::error_code error) -> void
        {
            if (!error)
                std::cout << "SUCCESS! Received message from: " << message.getVehicleId() << "\n";
        }, 3s);

    sleep(1);
    sender.send(PlatoonMessage::followerRequest(42), "127.0.0.1", 10000);
}

void test::testPeriodicTimer()
{
    Networking net;

    auto timer = time::Timer::create(net);

    int run = 0;
    std::atomic<bool> running{true};
    auto startTime = time::now();
    timer->startPeriodicTimeout(
        1s,
        [&run, &running, timer, &startTime]()
        {
            if (run >= 3)
            {
                std::cout << "SUCCESS!\n";
                timer->disable();
                running = false;
                return;
            }

            auto deltaTime = std::chrono::duration_cast<std::chrono::milliseconds>(time::now() - startTime);
            startTime = time::now();
            std::cout << "Delta time [ms]: " << deltaTime.count() << "\n";
            assert(std::abs(deltaTime) <= 2ms);
            run++;
        });

    while(running);
}

