//
// Created by philipp on 30.11.17.
//

#include <iostream>
#include "../include/LeaderVehicle.h"
#include "../include/Utils.h"

using namespace protocol;
using namespace std::chrono_literals;

constexpr networking::time::Duration LeaderVehicle::HEARTBEAT_TIMEOUT;
constexpr networking::time::Duration LeaderVehicle::BROADCAST_INTERVAL;

LeaderVehicle::LeaderVehicle(PrivateTag, networking::Networking & net, VehicleId vehicleId)
    : Vehicle(net, vehicleId, Role::LEADER)
      , broadcastTimer(networking::time::Timer::create(net))
      , broadcastSender(net)
{
    using namespace std::placeholders;
    platoonCreateServiceServer = networking::service::Server<PlatoonService>::create(
        net, "127.0.0.1", vehicleId + 10000,  // TODO: endpoint information!
        std::bind(&LeaderVehicle::makePlatoonCreateResponse, this, _1, _2));
}

LeaderVehicle::Ptr LeaderVehicle::create(networking::Networking & net, VehicleId vehicleId)
{
    return std::make_shared<LeaderVehicle>(PrivateTag{}, net, vehicleId);
}

void LeaderVehicle::doOnCreatePlatoon()
{
    log("leader[", std::to_string(vehicleId), "] starts creating platoon\n");
    // Generate a platoon id.
    platoonId = rand();

    platoonCreateServiceServer->advertiseService();

    scanAvailableVehicles();
    vehicleEndpointIter = availableVehicleEndpoints.begin();

    state = State::CREATING_PLATOON;
}

void LeaderVehicle::doCreatePlatoon()
{
    {
        std::lock_guard<std::mutex> lock(followersMutex);
        if (!followers.empty())
        {
            state = State::ON_RUNNING_PLATOON;
            return;
        }
    }

    auto vehicleEndpoint = *vehicleEndpointIter;
    makePlatoonCreateRequest(vehicleEndpoint);
    utils::cycle(vehicleEndpointIter, availableVehicleEndpoints);
}

void LeaderVehicle::doOnRunningPlatoon()
{
    log("leader[", std::to_string(vehicleId), "] starts running a platoon\n");

    startBroadcasting();

    state = State::RUNNING_PLATOON;
}

void LeaderVehicle::doRunningPlatoon()
{
    // Since broadcasting and heartbeat-receiving depends on timeouts (and therefore runs asynchronously)
    // we don't do anything here.
}

void LeaderVehicle::doOnLeavingPlatoon()
{
    stopBroadcasting();

    // Stop receiving heartbeats.
    {
        std::lock_guard<std::mutex> lock(followersMutex);
        for (auto & pair : followers)
        {
            // We don't remove the follower from the map since it might be the case
            // that there's a running handler which is currently using the follower object.
            auto & follower = pair.second;
            follower.heartbeatReceiver->stopReceiving();
        }
    }

    state = State::LEAVING_PLATOON;
}

void LeaderVehicle::doLeavingPlatoon()
{
    // Wait until each heartbeat-receive-handler has been dispatched and removed its corresponding follower.
    if (followers.empty())
        state = State::IDLE;
}

void LeaderVehicle::makePlatoonCreateRequest(VehicleEndpoint vehicleEndpoint)
{
    try
    {
        // We'll send a request indicating that we want to be the leader.
        auto request = PlatoonMessage::leaderRequest(vehicleId, platoonId);
        auto response = platoonCreateServiceClient.call(vehicleEndpoint.getHost(),
                                                        vehicleEndpoint.getTcpPort(),
                                                        request,
                                                        3s);
        // Verify response.
        if (response.getVehicleId() == vehicleEndpoint.getVehicleId() &&
            response.getMessageType() == messageTypes::ACCEPT_RESPONSE &&
            response.getPlatoonId() == platoonId)
        {
            addFollower(vehicleEndpoint);
        }
    }
    catch (const std::exception & e)
    {
        // This maybe a timeout or unable to connect to the other vehicle.
        // We just keep retrying in the next runs, so we don't need to handle this any further.
    }
}

PlatoonMessage LeaderVehicle::makePlatoonCreateResponse(const tcp::endpoint & clientEndpoint,
                                                        const PlatoonMessage & request)
{
    // Only consider this message if it is a follower request, meaning that a follower wants us to be its leader.
    if (request.getMessageType() != messageTypes::FOLLOWER_REQUEST)
        return PlatoonMessage::rejectResponse(vehicleId);

    auto followerId = request.getVehicleId();
    // TODO: change vehicle endpoint information!
    VehicleEndpoint followerEndpoint{
        followerId,
        clientEndpoint.address().to_string(),
        followerId + 10000,
        followerId + 10010};
    addFollower(followerEndpoint);

    return PlatoonMessage::acceptResponse(vehicleId, platoonId);
}

std::size_t LeaderVehicle::getNumFollowers()
{
    std::lock_guard<std::mutex> lock(followersMutex);
    return followers.size();
}

void LeaderVehicle::setPlatoonConfig(const PlatoonConfig & platoonConfig)
{
    std::lock_guard<std::mutex> lock(platoonConfigMutex);
    this->platoonConfig = platoonConfig;
}

bool LeaderVehicle::findFollower(VehicleId id)
{
    return followers.find(id) != followers.end();
}

void LeaderVehicle::addFollower(const VehicleEndpoint & endpoint)
{
    std::lock_guard<std::mutex> lock(followersMutex);
    if (findFollower(endpoint.getVehicleId()))
        return;

    Follower follower;
    follower.endpoint = endpoint;
    follower.heartbeatReceiver = networking::message::DatagramReceiver<PlatoonMessage>::create(
        net, endpoint.getHost(), endpoint.getUdpPort());
    follower.timestamp = networking::time::now();
    followers[endpoint.getVehicleId()] = follower;
    receiveHeartbeats(followers[endpoint.getVehicleId()]);
}

void LeaderVehicle::removeFollower(Follower & follower)
{
    follower.heartbeatReceiver->stopReceiving();
    std::lock_guard<std::mutex> lock(followersMutex);
    followers.erase(follower.endpoint.getVehicleId());
}

void LeaderVehicle::receiveHeartbeats(Follower & follower)
{
    auto self = shared_from_this();
    follower.heartbeatReceiver->asyncReceive(
        [self, &follower](const auto & senderEndpoint, const auto & message, const auto & error)
        {
            // Get the time passed since the last heart beat.
            auto timePassed = networking::time::now() - follower.timestamp;

            if (error == boost::asio::error::operation_aborted || timePassed >= HEARTBEAT_TIMEOUT)
            {
                // We remove the follower if the heartbeat times out or we were stopped.
                self->removeFollower(follower);
            }
            else if (error)
            {
                // In case of any other error we just ignore the message and receive again.
                self->receiveHeartbeats(follower);
            }
            else
            {
                // No error: check if we've got a leave-platoon-message.
                if (message.getMessageType() == messageTypes::LEAVE_PLATOON)
                {
                    self->removeFollower(follower);
                    return;
                }

                // In case of a heart beat message we set the timestamp to the current time to clear the heart beat timeout.
                if (message.getMessageType() == messageTypes::HEARTBEAT)
                    follower.timestamp = networking::time::now();

                self->receiveHeartbeats(follower);
            }
        }, HEARTBEAT_TIMEOUT);
}

void LeaderVehicle::startBroadcasting()
{
    auto self = shared_from_this();
    broadcastTimer = networking::time::Timer::create(net);
    broadcastTimer->startPeriodicTimeout(
        BROADCAST_INTERVAL,
        [self]()
        {
            std::lock_guard<std::mutex> lock(self->followersMutex);
            for (const auto & pair : self->followers)
            {
                const auto & follower = pair.second;
                auto broadcastMsg = PlatoonMessage::broadcastMessage(
                    follower.endpoint.getVehicleId(),
                    self->platoonId,
                    self->platoonConfig.platoonSpeed,
                    self->platoonConfig.innerPlatoonDistance);
                self->broadcastSender.send(broadcastMsg, follower.endpoint.getHost(), follower.endpoint.getUdpPort());
            }
        });
}

void LeaderVehicle::stopBroadcasting()
{
    broadcastTimer->disable();
}


